require 'virtus'

module ParaDice
  # @abstract  A module meant to be the basis of a custom Die object with
  #   behaviour included from this die.
  # @see GothamDice
  # @see StarWarsDice
  module Die
    ERROR_MSG = ': Must be implemented in your Die class or included in a faces module.'

    # @!scope class
    #   @!method new(params = {})
    #     @see Virtus for details
    #     creates new Die object and populates it's attributes with values from params.
    #       this list may not be comprehensive.  Check modules you include for other
    #       options.
    #     @param [Hash] params
    #       @option params [String] :name
    #       @option params [Random] :rng
    #       @option params [String] :description
    #     @return [Die]
    include Virtus.module

    # @!attribute [rw] name
    #   @return [String] required
    # @!attribute [rw] rng
    #   @return [Random] default Random.new
    # @!attribute [rw] description
    #   @return [String] default ''
    attribute :name, String, default: :quit_if_no_name
    attribute :rng, Random, default: ->(*args) { Random.new() }
    attribute :description, String, default: ''

    # for overloading and calling super on to allow more complex arrangements of
    #   face and dice behavior modules
    # @param [Random] roll_rng
    # @return [Object] random face from die
    def roll(roll_rng = rng)
      random_face(roll_rng)
    end

    # @abstract Implement in Include-able Module  and override {#face_count} to implement
    #   a custom Threadable class.
    def face_count
      raise NotImplementedError.new ('face_count' + ERROR_MSG)
    end

    # @abstract Implement in Include-able Module  and override {#face_count} to implement
    #   a custom Threadable class.
    def random_face(this_rng = rng)
      raise NotImplementedError.new ('random_face' + ERROR_MSG)
    end

    # @abstract Implement in Include-able Module  and override {#face_count} to implement
    #   a custom Threadable class.
    def faces
      raise NotImplementedError.new ('faces' + ERROR_MSG)
    end

    private
    def quit_if_no_name
      raise ArgumentError.new('no :name entry found in initialization hash')
    end
  end
end
